# this duplicates what the stateFlight code does, but returns a result
# saying whether the shot landed on a planet, whether it landed on a
# target, or if it was still in flight at the end. Also returns a
# closest point of approach and shortest distance (which will be zero
# for a successful shot).

# Takes an initial
# location, power, angle, and number of iterations.

# perhaps stateFlight could end up running through this code, too.


import math



GRAV_CONST = 800000
BASE_SPEED = 300.0



def evaluateLaunch(physicsObjectList,
                   power,
                   angle,
                   numIter,
                   launchTime,
                   timeStep,
                   sourcePlanet):

    # does a bunch of stuff

    startPlanetRad=sourcePlanet["radius"]
    px,py=sourcePlanet["center"]
    px+=math.cos(angle)*startPlanetRad
    py+=math.sin(angle)*startPlanetRad

    vx=math.cos(angle)*power*BASE_SPEED
    vy=math.sin(angle)*power*BASE_SPEED

    landedOnTarget=False
    landedOnPlanet=False
    stillInFlight=True
    closestPointOfApproach=(px,py)
    closestDistance=calcClosestDist(physicsObjectList,
                                    (px,py))
    pointList=[(px,py)]


    for step in range(numIter):
        (bLandedOnTarget,
         bLandedOnPlanet,
         bStillInFlight,
         p,
         newPoint,
         newVel)=evaluateTick(px,py,
                              vx,vy,
                              physicsObjectList,
                              launchTime+step*timeStep,
                              timeStep,
                              sourcePlanet)

        pointList.append(newPoint)
        
        d=calcClosestDist(physicsObjectList,
                          newPoint)

        if d<closestDistance:
            closestDistance=d
            closestPointOfApproach=newPoint

        px,py=newPoint
        vx,vy=newVel
        
        if not bStillInFlight:
            landedOnTarget=bLandedOnTarget
            landedOnPlanet=bLandedOnPlanet
            stillInFlight=False
            break
        

    

    return (landedOnTarget,
            landedOnPlanet,
            stillInFlight,
            closestPointOfApproach,
            closestDistance,
            pointList)


def calcClosestDist(physicsObjects,
                    (px,py)):

    bestDist=-1
    
    for p in physicsObjects:
        if not p.get("target",False):
            continue
        plX,plY=p["center"]
        r=p["radius"]

        dx=px-plX
        dy=py-plY

        d=math.sqrt(dx*dx+dy*dy)
        elev=d-r

        if (bestDist==-1 or
            elev<bestDist):
            bestDist=elev

    return bestDist



def evaluateTick(packageX, packageY,
                 packageVX, packageVY,
                 physicsObjects,
                 flightTime,
                 dt,
                 sourcePlanet): #in fractions of a second

    landedOnTarget=False
    landedOnPlanet=False
    stillInFlight=True

    ax,ay=calcAcceleration(packageX,
                           packageY,
                           physicsObjects,
                           flightTime)
    
    oldX=packageX
    oldY=packageY

    # use simplectic integrator:
    # the idea is to update velocity first, then position
    
    packageVX += ax*dt
    packageVY += ay*dt
    packageX += packageVX*dt
    packageY += packageVY*dt

        
    p=landed(packageX,
             packageY,
             oldX,
             oldY,
             sourcePlanet,
             physicsObjects,
             flightTime)

    if p:
        if p.get("target",False):
            landedOnTarget=True
            landedOnPlanet=True
            stillInFlight=False
        elif (p.get("gaseous", False) or
              p.get("dust", False)):
            stillInFlight=False
            landedOnPlanet=False
            landedOnTarget=False
        else:
            landedOnPlanet=True
            landedOnTarget=False
            stillInFlight=False
            

    newPoint=packageX,packageY
    newVel=packageVX,packageVY

    return (landedOnTarget,
            landedOnPlanet,
            stillInFlight,
            p,
            newPoint,
            newVel)


def calcAcceleration(px, py,
                     physicsObjects,
                     flightTime):
    # return ax,ay

    # Force = -gMM/r^2

    # factor out the package's mass to get acceleration:

    # a = -gM/r^2

    ax=0
    ay=0

    NO_ACCEL_PERIOD=0.5
    GRACE_PERIOD=1.0

    if flightTime<NO_ACCEL_PERIOD:
        return ax,ay

    for obj in physicsObjects:
        ox,oy=obj["center"]
        dx=ox-px
        dy=oy-py
        rSqr=dx*dx+dy*dy
        r=math.sqrt(rSqr)
        totalA=GRAV_CONST*obj["mass"]/rSqr

        oax=totalA*dx/r
        oay=totalA*dy/r

        ax+=oax
        ay+=oay

    if flightTime<GRACE_PERIOD:
        r=(flightTime-NO_ACCEL_PERIOD)/(GRACE_PERIOD-NO_ACCEL_PERIOD)
        ax*=r
        ay*=r
    return ax,ay
        
    


def landed(px, py, oldX, oldY, curPlan, physicsObjects, flightTime):
    for obj in physicsObjects:
        if obj==curPlan and flightTime<1.0:
            continue
        ox,oy=obj["center"]
        dx=ox-px
        dy=oy-py
        rSqr=dx*dx+dy*dy
        objRad=obj["radius"]

        if (rSqr<=objRad*objRad or
            penetrate(px, py, oldX, oldY, ox, oy, objRad)):
            return obj
    return None

def penetrate(p1x, p1y, p2x, p2y, px, py, rad):
    vx = p2x - p1x
    vy = p2y - p1y
    wx = px - p1x
    wy = py - p1y

    uNum = (px-p1x)*(p2x-p1x)+(py-p1y)*(p2y-p1y)

    dpx=p2x-p1x
    dpy=p2y-p1y

    uDenom=dpx*dpx+dpy*dpy

    u=uNum/uDenom

    if u<0 or u>1:
        return False

    pcx=p1x+u*dpx
    pcy=p1y+u*dpy

    dx=pcx-px
    dy=pcy-py

    distSqr=dx*dx+dy*dy
    radSqr=rad*rad

    return distSqr<radSqr
    

